/* Copyright 2022 NVIDIA Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "cunumeric/set/unique.h"
#include "cunumeric/set/unique_template.inl"

#include <omp.h>

namespace cunumeric {

using namespace Legion;
using namespace legate;

template <LegateTypeCode CODE, int32_t DIM>
struct UniqueImplBody<VariantKind::OMP, CODE, DIM> {
  using VAL = legate_type_of<CODE>;

  std::pair<Buffer<VAL>, size_t> operator()(const AccessorRO<VAL, DIM>& in,
                                            const Pitches<DIM - 1>& pitches,
                                            const Rect<DIM>& rect,
                                            const size_t volume,
                                            const std::vector<comm::Communicator>& comms,
                                            const DomainPoint& point,
                                            const Domain& launch_domain)
  {
    const auto max_threads = omp_get_max_threads();
    std::vector<std::set<VAL>> dedup_set(max_threads);

#pragma omp parallel
    {
      const int tid      = omp_get_thread_num();
      auto& my_dedup_set = dedup_set[tid];
#pragma omp for schedule(static)
      for (size_t idx = 0; idx < volume; ++idx) {
        auto p = pitches.unflatten(idx, rect.lo);
        my_dedup_set.insert(in[p]);
      }
    }

    size_t remaining = max_threads;
    size_t radix     = (max_threads + 1) / 2;
    while (remaining > 1) {
#pragma omp for schedule(static, 1)
      for (size_t idx = 0; idx < radix; ++idx) {
        if (idx + radix < remaining) {
          auto& my_set    = dedup_set[idx];
          auto& other_set = dedup_set[idx + radix];
          my_set.insert(other_set.begin(), other_set.end());
        }
      }
      remaining = radix;
      radix     = (radix + 1) / 2;
    }

    auto& final_dedup_set = dedup_set[0];
    size_t size           = final_dedup_set.size();
    size_t pos            = 0;
    auto kind   = CuNumeric::has_numamem ? Memory::Kind::SOCKET_MEM : Memory::Kind::SYSTEM_MEM;
    auto result = create_buffer<VAL>(size, kind);

    for (auto e : final_dedup_set) result[pos++] = e;

    return std::make_pair(result, size);
  }
};

/*static*/ void UniqueTask::omp_variant(TaskContext& context)
{
  unique_template<VariantKind::OMP>(context);
}

}  // namespace cunumeric
